# the bit that listens idly and serves RPCs
# and executor is the thing that takes a function, and executes it in a thread/process and returns the result as response or even another HTTP req entirely

from .utils import deserialize, serialize, get_self_ip
from .comms_wrappers import simplex_wrapper, duplex_wrapper
from .config import comms_config, default_rpc_config
from .inline_executor import InlineExecutor

from flask import Flask
from flask import request as route_req
from multiprocessing import Process
import threading
import inspect
import copy

# where the rpcs being offered are stored
rpcs = []
# the ip address of the worker
ip_addr = get_self_ip()
# the app that listens for incomming http requests
app = Flask(__name__)
name = 'worker'
node_type = 'worker'

# a default route to serve up what rpcs this worker offers, and their configuration
@app.route('/_get_profile', methods=['GET'])
def _get_profile():
	global name, ip_addr, rpcs

	profile = {
		'name': name,
		'ip_addr': ip_addr,
		'rpcs': rpcs
	}

	return serialize(profile)

# a default route to kill the worker in case a bug blocks SIGINT
@app.route('/kill', methods=['GET'])
def kill():
	func = route_req.environ.get('werkzeug.server.shutdown')

	if func is None:
		raise RuntimeError('Not running with the Werkzeug Server')

	func()
	return 'shutting down'

# a default route to provide basic info about the axon node, namely, that it's a worker
@app.route('/_type', methods=['GET'])
def _type():
	return node_type

# accepts two dicts, target and source
# in any shared keys between the two will be overwritten to source's value, and any keys in source will be copied to target, with thei values
def overwrite(target, source):
	# to avoid aliasing
	target = copy.copy(target)

	for key in source:
		target[key] = source[key]

	return target

def get_executor(configuration, fn):
	# wraps fn in two layers
	# the outer layer handles the communication pattern, being either simplex or duplex
	# the inner layer handles the execution environment, be that a thread or a process

	comms_pattern = configuration['comms_pattern']
	comms_wrapper = None
	if (comms_pattern == 'simplex'):
		comms_wrapper = simplex_wrapper
	elif (comms_pattern == 'duplex'):
		comms_wrapper = duplex_wrapper
	else:
		raise BaseException('unrecognized comms_pattern: '+str(comms_pattern))

	try:
		executor = configuration['executor']
	except(KeyError):
		raise(KeyError('executor not specified in RPC configuration'))

	if (executor == 'inline'):
		executor = InlineExecutor
	elif (executor == 'Process'):
		executor = Process
	elif((executor == 'Thread')):
		executor = threading.Thread
	else:
		raise(BaseException('unrecognized executor configuration:', executor))

	wrapped_fn = comms_wrapper(fn, executor)

	return wrapped_fn

# this function returns a function which will be applied to functions which the caller wishes to turn into RPCs
def rpc(**configuration):
	# accepts options for the RPC, simplex/duplex, and running environment
	# sets up the function that registers the route

	# overwrites the default configuration with keys from caller input
	configuration = overwrite(default_rpc_config, configuration)

	# this is the function that will be applied to functions which the caller wishes to turn into RPCs
	# accepts a function, which it will wrap in an executor function depending on the configuration passed to rcp, this wrapped function will be registered as a route
	def init_rpc(fn):
		name = fn.__name__ 

		rpc_desc = {
			'configuration': configuration,
			'name': name,
			'signature': inspect.signature(fn)
		}
		rpcs.append(rpc_desc)

		# wraps fn in the needed code to deserialize parameters from the request return its results, as well as run in a separate process/thread
		route_fn = get_executor(configuration, fn)
		# functions that flask registers as routes must have different names, even if they are registered at different endpoints. get_executor returns a function called 'wrapped_fn' every time, so we've got to change the name of the function here
		route_fn.__name__ = name
		# the endpoint of the route which exposes the function
		endpoint = '/'+configuration['endpoint_prefix']+name

		# registering the function as a route
		app.route(endpoint, methods=['POST'])(route_fn)

		return fn

	return init_rpc

# starts the web app
def init(wrkr_name='worker'):
	global name

	name = wrkr_name
	# the web application that will serve the rpcs as routes
	app.run(host='0.0.0.0', port=comms_config.worker_port, threaded=False)


from .config import comms_config
from .utils import deserialize, serialize, POST
from .inline_executor import InlineExecutor, inline_lock

from flask import Flask
from flask import request as route_req
from multiprocessing import Process, Manager
from threading import Thread
import threading
import requests
import inspect
import sys
import asyncio
import traceback

# catches any errors in fn, and handles them properly
def error_wrapper(fn):

	def wrapped_fn(args, kwargs):
		return_object = {
			'errcode': 0,
			'result': None,
		}

		try:
			# execute the given function
			return_object['result'] = fn(*args, **kwargs)

		except:

			return_object['errcode'] = 1
			return_object['result'] = (traceback.format_exc(), sys.exc_info()[1])

		return return_object

	return wrapped_fn

def async_wrapper(fn):

	def wrapped_fn(params):
		return asyncio.run(fn(params))

	return wrapped_fn

# wraps fn in the needed for simplex communication pattern
# adds code to deserialize parameters from the request and return its results through the response
def simplex_wrapper(fn, executor):

	def wrapped_fn():

		mp_manager = Manager()
		result_holder = mp_manager.dict()

		# this function will run in a separate thread, will execute the RPC and return the result with a POST request
		def wrk_fn(fn, params_str, result_holder):

			# deserializes parameters
			args, kwargs = deserialize(params_str)

			# if fn is a coroutine, it needs to be run with an event loop
			if inspect.iscoroutinefunction(fn): fn = async_wrapper(fn)

			# catches and returns any error
			fn = error_wrapper(fn)

			# runs the function
			return_object = fn(args, kwargs)

			result_holder['result'] = serialize(return_object)

		params_str = route_req.form['msg']

		fn_executor = executor(target=wrk_fn, args=(fn, params_str, result_holder))
		fn_executor.deamon = True
		fn_executor.start()

		fn_executor.join()

		return result_holder['result']

	return wrapped_fn



# wraps fn in the needed for duplex communication pattern
# adds code to deserialize parameters from the request and return its results through the a separate request
def duplex_wrapper(fn, executor):

	def wrapped_fn():

		# this function will run in a separate thread, will execute the RPC and return the result with a POST request
		def wrk_fn(fn, params_str, calling_ip):

			# deserializes parameters
			(call_info, args, kwargs) = deserialize(params_str)

			# the info needed to return the result
			call_id = call_info['id']
			rvl_port = call_info['rvl_port']

			# if fn is a coroutine, it needs to be run with an event loop
			if inspect.iscoroutinefunction(fn): fn = async_wrapper(fn)

			# catches and returns any error
			fn = error_wrapper(fn)

			# runs the function
			return_object = fn(args, kwargs)

			# returns the result via a POST request
			url = 'http://'+calling_ip+':'+str(rvl_port)+'/_return_value_linker'
			data = {'msg': serialize((call_id, return_object))}
			requests.post(url=url, data=data)


		params_str = route_req.form['msg']
		calling_ip = route_req.remote_addr

		# because flask, and WSGI servers in general, cannot schedule code to execute after a route completes, inline duplex RPCs would have to send the returning request before the calling request completes
		# to prevent this, we must run the RPC in a separate thread, and to preserve synchronous bahavior, we must use a threadlock with other inline RPCs. This will ensure no two inline RPCs are executing at the same time.
		fn_executor = None
		if(executor == InlineExecutor):

			def inline_lock_fn(fn, params_str, calling_ip):
				with inline_lock:
					result = wrk_fn(fn, params_str, calling_ip)

				return result

			fn_executor = Thread(target=inline_lock_fn, args=(fn, params_str, calling_ip))
		else:
			fn_executor = executor(target=wrk_fn, args=(fn, params_str, calling_ip))

		fn_executor.deamon = True
		fn_executor.start()

		return serialize('duplex')

	return wrapped_fn


from types import SimpleNamespace

default_rpc_config = {
	'comms_pattern': 'simplex',
	'endpoint_prefix': 'rpc/',
	'executor': 'inline'
}

comms_config = {
	'notice_board_port': 8002,
	'request_timeout': 30,
	'RVL_port': 8001,
	'worker_port': 8000
}

comms_config = SimpleNamespace(**comms_config)


from .config import comms_config

import pickle
import requests
import aiohttp
import asyncio
import codecs
import time
import socket

def GET(url, timeout=comms_config.request_timeout):
	resp = requests.get(url, timeout=timeout)
	return resp.status_code, resp.text

async def async_GET(url, timeout=comms_config.request_timeout):
	async with aiohttp.ClientSession(conn_timeout=timeout) as session:
		async with session.get(url) as resp:
			return resp.status, await resp.text()

def POST(url, data=None, timeout=comms_config.request_timeout):
	resp = requests.post(url=url, data=data, timeout=timeout)
	return resp.status_code, resp.text

async def async_POST(url, data=None, timeout=comms_config.request_timeout):
	async with aiohttp.ClientSession(conn_timeout=timeout) as session:
		async with session.post(url, data=data) as resp:
			return resp.status, await resp.text()

# pickle operates on bytes, but http operates on strings, so we've gotta convert pickles to and from a string
def serialize(obj):
	pickled = pickle.dumps(obj)
	return codecs.encode(pickled, "base64").decode()

# pickle operates on bytes, but http operates on strings, so we've gotta convert pickles to and from a string
def deserialize(obj_str):
	obj_bytes = codecs.decode(obj_str.encode(), "base64")
	return pickle.loads(obj_bytes)

def get_active_workers():
	status, text = GET('http://'+str(comms_config.notice_board_ip)+':'+str(comms_config.notice_board_port)+'/get_ips')

	if (status == 404): print('notice board not found')

	return deserialize(text)

self_ip = None
def get_self_ip():
	global self_ip

	if (self_ip == None):
		s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
		s.connect(("8.8.8.8", 80))
		ip = s.getsockname()[0]
		s.close()

		self_ip = ip

		return ip

	else:
		return self_ip

# scans for an available port in between upper_bound and lower_bound
def get_open_port(lower_bound=8000, upper_bound=9000):

	for target_port in range(8000, upper_bound):
		# defines a socket, AF_INIT means ipv4 and SOCK_STREAM means TCP
		sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		location = ("127.0.0.1", target_port)
		# will return 0 if the port is available
		result = sock.connect_ex(location)

		if result != 0:
			sock.close()
			return target_port

	sock.close()
	raise BaseException('No available ports in between '+str(lower_bound)+' and '+str(upper_bound))


# the bit that exports worker profile class and uses the RVL
# an RPC stub is the thing on the client that makes a calling request and waits for the response

from .utils import serialize, deserialize, GET, POST, async_GET, async_POST
from .config import comms_config, default_rpc_config
from .return_value_linker import ReturnEvent, RVL

from types import SimpleNamespace
import asyncio
import types
import uuid

rvl = None

async def start_client():
	global rvl

	loop = asyncio.get_running_loop()
	rvl = RVL(loop)
	rvl.start()

def get_worker_profile(ip_addr):
	url = 'http://'+str(ip_addr)+':'+str(comms_config.worker_port)+'/_get_profile'
	_, profile_str = GET(url)
	return deserialize(profile_str)

def error_handler(worker_ip, return_obj):
	if (return_obj['errcode'] == 1):
		# an error occured in worker, raise it
		(error_info, error) = return_obj['result']

		print('the following error occured in worker at '+str(worker_ip)+':')
		print(error_info)
		raise(error)

	else:
		# returns the result
		return return_obj['result']

def get_simplex_rpc_stub(worker_ip, rpc_name):

	# URL of RPC
	url = 'http://'+str(worker_ip)+':'+str(comms_config.worker_port)+'/'+default_rpc_config['endpoint_prefix']+rpc_name

	# this function makes a calling request to a simplex RPC on the worker with IP worker_ip to the RPC named rpc_name
	async def simplex_rpc_stub(*args, **kwargs):
		
		# makes the calling request
		status, text = await async_POST(url=url , data={'msg': serialize((args, kwargs))})

		if (text =='duplex'):
			raise(BaseException('simplex call sent to duplex RPC'))

		# deserializes return object from worker
		return_obj = deserialize(text)
		return error_handler(worker_ip, return_obj)

	return simplex_rpc_stub

def get_duplex_rpc_stub(worker_ip, rpc_name):

	# the URL of the RPC
	url = 'http://'+str(worker_ip)+':'+str(comms_config.worker_port)+'/'+default_rpc_config['endpoint_prefix']+rpc_name

	# this function makes a calling request to a simplex RPC on the worker with IP worker_ip to the RPC named rpc_name
	async def duplex_rpc_stub(*args, **kwargs):
		global rvl
		if rvl == None:
			await start_client()

		# we must register an event listenner with the return value linker, which will wait for the incomming result request
		# this uuid will identify the call, so the RVL can lookup the event
		call_id = uuid.uuid4()
		return_event = ReturnEvent()
		rvl.register(call_id, return_event)

		# this object holds information about the function call that the worker will need to provide to the rvl upon completion
		call_info = {'id': call_id, 'rvl_port': rvl.port}

		# makes the calling request
		status, text = await async_POST(url=url , data={'msg': serialize((call_info, args, kwargs))})
		return_obj = deserialize(text)

		# raises exception if the receiving RPC is simplex, not duplex
		if (return_obj != 'duplex'):
			print(return_obj)
			raise(BaseException('duplex call sent to simplex RPC'))

		# we now wait for the result to return
		return_obj = await return_event.get_return_value()

		# detects if an error was thrown in worker
		return error_handler(worker_ip, return_obj)

	return duplex_rpc_stub

class RemoteWorker():

	def __init__(self, profile_or_ip):
		self.ip_addr = None
		self.rpcs = None

		profile = None

		# profile_or_ip is either an ip address or a worker profile, here we test to see which one
		try:
			profile_or_ip['ip_addr']
			# profile_or_ip has a member ip_addr, and so must be a profile
			profile = profile_or_ip

		except(TypeError):
			# profile_or_ip is an ip address
			profile = get_worker_profile(profile_or_ip)

		# sets up all the rfc stubs
		self.setup(profile)

	def setup(self, profile):
		self.ip_addr = profile['ip_addr']
		self.name = profile['name']

		self.setup_rpc_stubs(profile['rpcs'])

	def setup_rpc_stubs(self, rpcs_profiles):
		rpcs = {}

		for rpc_desc in rpcs_profiles:
			
			name = rpc_desc['name']
			configuration = rpc_desc['configuration']

			if (configuration['comms_pattern'] == 'simplex'):
				rpcs[name] = get_simplex_rpc_stub(self.ip_addr, name)

			elif (configuration['comms_pattern'] == 'duplex'):
				rpcs[name] = get_duplex_rpc_stub(self.ip_addr, name)
			else:
				raise BaseException('unrecognised communication pattern:'+str(configuration['comms_pattern']))

		self.rpcs = SimpleNamespace(**rpcs)


import pathlib
from setuptools import setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
	name="axon-ECRG",
	version="0.1.3",
	description="Edge computing framework developed and maintained by the Edge Computing Research Group at Queen's University",
	long_description=README,
	long_description_content_type="text/markdown",
	url="https://github.com/DuncanMays/axon-ECRG",
	packages=['axon'],
	install_requires=['flask', 'uuid', 'requests', 'aiohttp'],
)
